// VBoxService.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "VBoxAutoStartSvc.h"


#define SVCNAME TEXT("VBoxAutoStart")
#define PRORGRAMFILES_PATH	(TEXT("%ProgramFiles%/Sun xVM VirtualBox/"))

SERVICE_STATUS          gSvcStatus; 
SERVICE_STATUS_HANDLE   gSvcStatusHandle; 
HANDLE                  ghSvcStopEvent = NULL;

VOID SvcInstall(TCHAR *user=NULL,TCHAR *pass=NULL);
VOID WINAPI SvcCtrlHandler( DWORD ); 
VOID WINAPI SvcMain( DWORD, LPTSTR * ); 

VOID ReportSvcStatus( DWORD, DWORD, DWORD );
VOID SvcInit( DWORD, LPTSTR * ); 
VOID SvcReportEvent( LPTSTR );
int  SvcQuery( );
int VMStart();

/*****************************************************************
*		VM Functions
******************************************************************/
//
// Purpose: 
//   Start all machines which have !EnableAutoStart sting in description
//
// Parameters:
//   None
// 
// Return value:
//   0: Error, 1:Success
//
int VMStart()
{
	wchar_t szExePath[_MAX_PATH];//VBoxHeadless path
	CComBSTR	bstrHomePath;
	HRESULT hr;
	unsigned long i,max_i;
	
	CComPtr<IVirtualBox> vbox;
	CComPtr<IMachineCollection> machines;

	//Get Virtualbox install dir
	CRegKey key;
	if( ERROR_SUCCESS != key.Open(HKEY_LOCAL_MACHINE,L"SOFTWARE\\Sun\\xVM VirtualBox",KEY_READ ) ) return 0; // on error exit
	ULONG ulen = MAX_PATH-1;
	if( ERROR_SUCCESS != key.QueryStringValue(L"InstallDir",szExePath,&ulen) ) return 0; 
	key.Close();
	
	wcsncat(szExePath,L"VBoxHeadless.exe",_MAX_PATH);

	hr = vbox.CoCreateInstance(CLSID_VirtualBox);
	if(S_OK != hr) return 0;
	hr = vbox->get_Machines(&machines);
	if(S_OK == hr) hr = machines->get_Count(&max_i);
	if(S_OK != hr) return 0;

	for( i=0;i<max_i;i++) {
		CComPtr<IMachine> machine;
		hr = machines->GetItemAt(i,&machine); 
		if(S_OK != hr) continue;
		CComBSTR temp;
		hr = machine->get_Description(&temp);
		if( hr != S_OK ||NULL == temp || NULL == wcsstr(temp,L"!EnableAutoStart") ) continue;
		
		CComBSTR name;

		//construct command line
		hr = machine->get_Name(&name);	
		temp = TEXT(" -v config -s ");
		temp += name;	
		
		//create process
		STARTUPINFO si;
		memset(&si,0,sizeof(si));
		si.cb = sizeof(si);
		PROCESS_INFORMATION pi; 
		int ret = CreateProcess(szExePath,temp,0,0,0,
			CREATE_NEW_PROCESS_GROUP|DETACHED_PROCESS,
			0,0,&si,&pi);
		ret = GetLastError();
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);

		
	}
	return 1;
}
//
// Purpose: 
//   Stop all running VM's which have !EnableAutoStart sting in description
//
// Parameters:
//   None
// 
// Return value:
//   0: Error, 1:Success
//
int VMStop()
{
	CComPtr<IVirtualBox> vbox;
	CComPtr<IMachineCollection> machines;
	HRESULT hr;
	unsigned long i,max_i;

	hr = vbox.CoCreateInstance(CLSID_VirtualBox);
	if(S_OK != hr) return 0;
	hr = vbox->get_Machines(&machines);
	if(S_OK == hr) hr = machines->get_Count(&max_i);
	if(S_OK != hr) return 0;
	for( i=0;i<max_i;i++) {
		CComPtr<IMachine>	machine;
		MachineState		vmstate;
		
		hr = machines->GetItemAt(i,&machine); 
		if(S_OK != hr) continue;
		
		
		//check for autostart flag
		CComBSTR temp;
		hr = machine->get_Description(&temp);
		if( hr != S_OK ||NULL == temp || NULL == wcsstr(temp,L"!EnableAutoStart") ) continue;
		
		//check for running state
		machine->get_State(&vmstate);
		if( MachineState_Running == vmstate || MachineState_Starting == vmstate ) {
			GUID vmID;
			CComPtr<ISession>   sess;
			CComPtr<IConsole>   console;
			hr	= machine->get_Id(&vmID);
			hr  = sess.CoCreateInstance(CLSID_Session);
			if(S_OK == hr ) hr = vbox->OpenExistingSession(sess,vmID);
			if(S_OK == hr ) hr = sess->get_Console(&console);
			if(S_OK == hr ) hr = console->PowerButton();
			if(S_OK == hr ) hr = sess->Close();
		}
	}
	return 1;
}
/*****************************************************************
*		Service functions - modified example from Windows SDK
******************************************************************/
//
// Purpose: 
//   Query service status
//
// Parameters:
//   None
// 
// Return value:
//   Service State: SERVICE_START_PENDING, . . . 
//
int SvcQuery()
{
    int nResult = NULL;

	SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, GENERIC_READ);

    if (hSCM != NULL)
    {
        SC_HANDLE hService = ::OpenService(hSCM, SVCNAME, SERVICE_QUERY_CONFIG|SERVICE_QUERY_STATUS);
        if (hService != NULL)
        {
			SERVICE_STATUS ss;
			BOOL nQs = QueryServiceStatus(hService,&ss);
           
			if( nQs ) 
				nResult = ss.dwCurrentState ;
			else 
				nResult = SERVICE_START_PENDING;
            
			::CloseServiceHandle(hService);
        }
        ::CloseServiceHandle(hSCM);
    }
    return nResult;
}
int SvcRemove()
{
    if (SvcQuery() == 0)
    return TRUE;

    SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

    if (hSCM == NULL)
    {
        fprintf(stderr,"Couldn't open service manager");
        return FALSE;
    }

    SC_HANDLE hService = ::OpenService(hSCM, SVCNAME, SERVICE_STOP | DELETE);

    if (hService == NULL)
    {
        ::CloseServiceHandle(hSCM);
        fprintf(stderr,"Couldn't open service");
        return FALSE;
    }
    SERVICE_STATUS status;
    ::ControlService(hService, SERVICE_CONTROL_STOP, &status);

    BOOL bDelete = ::DeleteService(hService);
    ::CloseServiceHandle(hService);
    ::CloseServiceHandle(hSCM);

    if (bDelete)
        return TRUE;

    fprintf(stderr,"Service could not be deleted");
    return FALSE;

	
}
//
// Purpose: 
//   Installs a service in the SCM database
//
// Parameters:
//   None
// 
// Return value:
//   None
//
VOID SvcInstall(TCHAR *user,TCHAR *pass)
{
    SC_HANDLE schSCManager;
    SC_HANDLE schService;
    TCHAR szPath[MAX_PATH];
	
	if(NULL != SvcQuery()) 
        return;
    
	if( !GetModuleFileName( NULL, szPath, MAX_PATH ) )
    {
        printf("Cannot install service (%d)\n", GetLastError());
        return;
    }
	
	//check windows version
	OSVERSIONINFO ov;
	ov.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	::GetVersionEx(&ov);
	BOOL isXP = //
				ov.dwMajorVersion  >= 5					&& 
				VER_PLATFORM_WIN32_NT == ov.dwPlatformId	&&
				ov.dwMinorVersion > 0	
				? 1 :0;

    // Get a handle to the SCM database. 
 
    schSCManager = OpenSCManager( 
        NULL,                    // local computer
        NULL,                    // ServicesActive database 
        SC_MANAGER_ALL_ACCESS);  // full access rights 
 
    if (NULL == schSCManager) 
    {
        printf("OpenSCManager failed (%d)\n", GetLastError());
        return;
    }

    // Create the service

    schService = CreateService( 
        schSCManager,              // SCM database 
        SVCNAME,                   // name of service 
        SVCNAME,                   // service name to display 
        SERVICE_ALL_ACCESS,        // desired access 
        SERVICE_WIN32_OWN_PROCESS, // service type 
        SERVICE_DEMAND_START,      // start type 
        SERVICE_ERROR_NORMAL,      // error control type 
        szPath,                    // path to service's binary 
        NULL,                      // no load ordering group 
        NULL,                      // no tag identifier 
        NULL,                      // no dependencies 
        isXP && user == NULL ? TEXT("NT AUTHORITY\\NetworkService") :user,    //account 
        isXP && user == NULL ? TEXT("") : pass);                     // password 
 
    if (schService == NULL) 
    {
        printf("CreateService failed (%d)\n", GetLastError()); 
        CloseServiceHandle(schSCManager);
        return;
    }
    else printf("Service installed successfully\n"); 

    CloseServiceHandle(schService); 
    CloseServiceHandle(schSCManager);
}

//
// Purpose: 
//   Entry point for the service
//
// Parameters:
//   dwArgc - Number of arguments in the lpszArgv array
//   lpszArgv - Array of strings. The first string is the name of
//     the service and subsequent strings are passed by the process
//     that called the StartService function to start the service.
// 
// Return value:
//   None.
//
VOID WINAPI SvcMain( DWORD dwArgc, LPTSTR *lpszArgv )
{
    // Register the handler function for the service

    gSvcStatusHandle = RegisterServiceCtrlHandler( 
        SVCNAME, 
        SvcCtrlHandler);

    if( !gSvcStatusHandle )
    { 
        SvcReportEvent(TEXT("RegisterServiceCtrlHandler")); 
        return; 
    } 

    // These SERVICE_STATUS members remain as set here

    gSvcStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS; 
    gSvcStatus.dwServiceSpecificExitCode = 0;    

    // Report initial status to the SCM

    ReportSvcStatus( SERVICE_START_PENDING, NO_ERROR, 3000 );

    // Perform service-specific initialization and work.

    SvcInit( dwArgc, lpszArgv );
}

//
// Purpose: 
//   The service code
//
// Parameters:
//   dwArgc - Number of arguments in the lpszArgv array
//   lpszArgv - Array of strings. The first string is the name of
//     the service and subsequent strings are passed by the process
//     that called the StartService function to start the service.
// 
// Return value:
//   None
//
VOID SvcInit( DWORD dwArgc, LPTSTR *lpszArgv)
{
    // TO_DO: Declare and set any required variables.
    //   Be sure to periodically call ReportSvcStatus() with 
    //   SERVICE_START_PENDING. If initialization fails, call
    //   ReportSvcStatus with SERVICE_STOPPED.

    // Create an event. The control handler function, SvcCtrlHandler,
    // signals this event when it receives the stop control code.

    ghSvcStopEvent = CreateEvent(
                         NULL,    // default security attributes
                         TRUE,    // manual reset event
                         FALSE,   // not signaled
                         NULL);   // no name

    if ( ghSvcStopEvent == NULL)
    {
        ReportSvcStatus( SERVICE_STOPPED, NO_ERROR, 0 );
        return;
    }

    // Report running status when initialization is complete.

    ReportSvcStatus( SERVICE_RUNNING, NO_ERROR, 0 );

    // TO_DO: Perform work until service stops.COINIT_APARTMENTTHREADED COINIT_MULTITHREADED
	CoInitializeEx(0,COINIT_MULTITHREADED );
	VMStart();

    while(1)
    {
        // Check whether to stop the service.
        WaitForSingleObject(ghSvcStopEvent, INFINITE);

        ReportSvcStatus( SERVICE_STOP_PENDING, NO_ERROR, 0 );
        break; 
    }
	VMStop();
	
	//wait for VM's to stop
	for( int i=0;i<20;i++ ) {
		ReportSvcStatus( SERVICE_STOP_PENDING, NO_ERROR, 0 );
		Sleep(500);
	}
	CoUninitialize();

	ReportSvcStatus( SERVICE_STOPPED, NO_ERROR, 0 );
}

//
// Purpose: 
//   Sets the current service status and reports it to the SCM.
//
// Parameters:
//   dwCurrentState - The current state (see SERVICE_STATUS)
//   dwWin32ExitCode - The system error code
//   dwWaitHint - Estimated time for pending operation, 
//     in milliseconds
// 
// Return value:
//   None
//
VOID ReportSvcStatus( DWORD dwCurrentState,
                      DWORD dwWin32ExitCode,
                      DWORD dwWaitHint)
{
    static DWORD dwCheckPoint = 1;

    // Fill in the SERVICE_STATUS structure.

    gSvcStatus.dwCurrentState = dwCurrentState;
    gSvcStatus.dwWin32ExitCode = dwWin32ExitCode;
    gSvcStatus.dwWaitHint = dwWaitHint;

    if (dwCurrentState == SERVICE_START_PENDING)
        gSvcStatus.dwControlsAccepted = 0;
    else gSvcStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;

    if ( (dwCurrentState == SERVICE_RUNNING) ||
           (dwCurrentState == SERVICE_STOPPED) )
        gSvcStatus.dwCheckPoint = 0;
    else gSvcStatus.dwCheckPoint = dwCheckPoint++;

    // Report the status of the service to the SCM.
    SetServiceStatus( gSvcStatusHandle, &gSvcStatus );
}

//
// Purpose: 
//   Called by SCM whenever a control code is sent to the service
//   using the ControlService function.
//
// Parameters:
//   dwCtrl - control code
// 
// Return value:
//   None
//
VOID WINAPI SvcCtrlHandler( DWORD dwCtrl )
{
   // Handle the requested control code. 

   switch(dwCtrl) 
   {  
	  case SERVICE_CONTROL_SHUTDOWN: 
      case SERVICE_CONTROL_STOP: 
         ReportSvcStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);

         // Signal the service to stop.

         SetEvent(ghSvcStopEvent);
         
         return;
 
      case SERVICE_CONTROL_INTERROGATE: 
         // Fall through to send current status.
         break; 
 
      default: 
         break;
   } 

   ReportSvcStatus(gSvcStatus.dwCurrentState, NO_ERROR, 0);
}

//
// Purpose: 
//   Logs messages to the event log
//
// Parameters:
//   szFunction - name of function that failed
// 
// Return value:
//   None
//
// Remarks:
//   The service must have an entry in the Application event log.
//
VOID SvcReportEvent(LPTSTR szFunction) 
{ 
    HANDLE hEventSource;
    LPCTSTR lpszStrings[2];
    TCHAR Buffer[80];

    hEventSource = RegisterEventSource(NULL, SVCNAME);

    if( NULL != hEventSource )
    {
        StringCchPrintf(Buffer, 80, TEXT("%s failed with %d"), szFunction, GetLastError());

        lpszStrings[0] = SVCNAME;
        lpszStrings[1] = Buffer;

        ReportEvent(hEventSource,        // event log handle
                    EVENTLOG_ERROR_TYPE, // event type
                    0,                   // event category
                    SVC_ERROR,           // event identifier
                    NULL,                // no security identifier
                    2,                   // size of lpszStrings array
                    0,                   // no binary data
                    lpszStrings,         // array of strings
                    NULL);               // no binary data

        DeregisterEventSource(hEventSource);
    }
}
//
// Purpose: 
//   Entry point for the process
//
// Parameters:
//   None
// 
// Return value:
//   None
//
int __cdecl _tmain(int argc, TCHAR *argv[]) 
{ //_asm int 3;
	//init runtime library locale functions
	setlocale( LC_ALL, "" );
    
	// TO_DO: Add any additional services for the process to this table.
    SERVICE_TABLE_ENTRY DispatchTable[] = 
    { 
        { SVCNAME, (LPSERVICE_MAIN_FUNCTION) SvcMain }, 
        { NULL, NULL } 
    };  
	//query service status
	int nServicestartupStatus = SvcQuery();
	if( nServicestartupStatus == SERVICE_START_PENDING ) {
		// This call returns when the service has stopped. 
		// The process should simply terminate when the call returns.
		if (!StartServiceCtrlDispatcher( DispatchTable )) 
		{ 
			SvcReportEvent(TEXT("StartServiceCtrlDispatcher")); 
		} 	
		return 0;
	} 

	// If command-line parameter is "install", install the service. 
    // Otherwise, the service is probably being started by the SCM.
	
    if( argc > 1 )
    {
		if( lstrcmpi( argv[1], TEXT("-install")) == 0 ) {
			if( argc > 3 ) {
				SvcInstall(argv[2],argv[3]);
			} else  {
				SvcInstall();
			}
		}
		if( lstrcmpi( argv[1], TEXT("-remove")) == 0 ) {
			SvcRemove();
		}
        return 0;
	} else {
		_tprintf(
			TEXT("\nUsage:	VBoxAutoStartSvc -install [<username>] [<[password>]\n")
			TEXT("        VBoxAutoStartSvc -remove\n"));
	}
#if 0
	CoInitializeEx(0,COINIT_MULTITHREADED );
	VMStart();
	VMStop();
    CoUninitialize();
#endif
} 
